
#include "stdafx.h"
#include "AString.h"
#include "FilePartition.h"
#include "DiffEngine.h"


// BuffBlock implementation ///////////////////////////////////////////////////

const int defaultallocationlength = 32768;

BuffBlock::BuffBlock()
{
    setDefaultAllocationLength(defaultallocationlength);
}

BuffBlock::~BuffBlock()
{
    free();
}

void BuffBlock::free()
{
    // free all blocks
    long nbBlocks = m_arrayBlockPtrs.GetSize();
    for (long i = 0; i < nbBlocks; i++)
    {
        void* buffer = m_arrayBlockPtrs.GetAt(i).m_pBuffer;
        if (buffer)
            ::free(buffer);
    }
    m_arrayBlockPtrs.RemoveAll();
}

void BuffBlock::setDefaultAllocationLength(long nLength) // in bytes
{
    m_nLengthNewBlock = nLength;
}

void* BuffBlock::getFirstBuffer() // pointer to first block
{
    if (m_arrayBlockPtrs.GetSize() == 0)
        return NULL;

    return m_arrayBlockPtrs.GetAt(0).m_pBuffer;
}

long BuffBlock::getFirstLength()
{
    if (m_arrayBlockPtrs.GetSize() == 0)
        return 0;

    return m_arrayBlockPtrs.GetAt(0).m_nLength;
}

long BuffBlock::getFirstAllocatedLength()
{
    if (m_arrayBlockPtrs.GetSize() == 0)
        return 0;

    return m_arrayBlockPtrs.GetAt(0).m_nAllocatedLength;
}

long BuffBlock::getTotalLength()
{
    long nTotalLength = 0;
    long nbBlocks = m_arrayBlockPtrs.GetSize();
    for (long i = 0; i < nbBlocks; i++)
        nTotalLength += m_arrayBlockPtrs.GetAt(i).m_nLength;
    return nTotalLength;
}

void* BuffBlock::getLastBuffer() // pointer to last block
{
    if (m_arrayBlockPtrs.GetSize() == 0)
        return NULL;

    return m_arrayBlockPtrs.GetAt(m_arrayBlockPtrs.GetSize() - 1).m_pBuffer;
}

long BuffBlock::getLastLength()
{
    if (m_arrayBlockPtrs.GetSize() == 0)
        return 0;

    return m_arrayBlockPtrs.GetAt(m_arrayBlockPtrs.GetSize() - 1).m_nLength;
}

long BuffBlock::getLastAllocatedLength()
{
    if (m_arrayBlockPtrs.GetSize() == 0)
        return 0;

    return m_arrayBlockPtrs.GetAt(m_arrayBlockPtrs.GetSize() - 1).m_nAllocatedLength;
}


long BuffBlock::getBlockCount()
{
    return m_arrayBlockPtrs.GetSize();
}

void* BuffBlock::getBlock(long i, long& o_nLength, long& o_nAllocatedLength)
{
    if (i < 0 || i >= getBlockCount())
        return NULL;

    BuffBlockUnit block = m_arrayBlockPtrs.GetAt(i);
    o_nLength = block.m_nLength;
    o_nAllocatedLength = block.m_nAllocatedLength;
    return block.m_pBuffer;
}

bool BuffBlock::addChar(char c)
{
    return addBuffer( &c, 1 );
}

bool BuffBlock::addEOLIfRequired()
{
    char* p = (char*) getLastBuffer();
    if (!p)
        return addChar(0);
    long len = getLastLength();
    if (len == 0)
        return addChar(0);
    if (p[len] != 0)
        return addChar(0);
    return true;
}

bool BuffBlock::addByte(unsigned char c)
{
    return addBuffer( &c, 1 );
}

bool BuffBlock::addWord(WORD w)
{
    return addBuffer( &w, 2 );
}

bool BuffBlock::addBuffer(/*in*/CString& s)
{
    return addBuffer((LPSTR)s.GetBuffer(0), s.GetLength());
}

bool BuffBlock::addBuffer(/*in*/LPSTR s)
{
    if (!s)
        return false;
    return addBuffer(s, strlen(s));
}

bool BuffBlock::addBuffer(/*in*/void* p, long length)
{
    m_pInsertionBuffer = NULL;

    if (!p || length == 0)
        return true;

    //do we need to create a new block?
    if (getLastLength() + length > getLastAllocatedLength())
    {
        long nLengthToAllocate = max(length, m_nLengthNewBlock);
        // new buffer of arbitrary size
        char* pNewBuffer = (char*) malloc(1 + nLengthToAllocate);
        if (!pNewBuffer)
            return false;

        BuffBlockUnit block;
        block.m_pBuffer = pNewBuffer;
        block.m_nLength = 0;
        block.m_nAllocatedLength = nLengthToAllocate;

        if (block.m_pBuffer == NULL ||
            block.m_nAllocatedLength == 0)
            return false;
    
        m_arrayBlockPtrs.Add(block);
    }


    BuffBlockUnit block;
    block = m_arrayBlockPtrs.GetAt( m_arrayBlockPtrs.GetSize() - 1 );

    if (!block.m_pBuffer)
        return false;

    m_pInsertionBuffer = ((char*)block.m_pBuffer) + block.m_nLength;

	// copy [in] buffer (size = nLength bytes)
	memcpy ( ((char*)block.m_pBuffer) + block.m_nLength, p, length);
	block.m_nLength += length;
	((char*)block.m_pBuffer)[block.m_nLength] = 0;

    // update the block in the array
    m_arrayBlockPtrs.SetAt(m_arrayBlockPtrs.GetSize() - 1, block);

    return true;
}

bool BuffBlock::addBuffer(/*in*/BuffBlock* in)
{
    if (!in)
        return false;

    long nbBlocks = in->getBlockCount();
    for (long iBlock = 0; iBlock < nbBlocks; iBlock++)
    {
        long n, nAllocatedLength;
        void* block = in->getBlock(iBlock, n, nAllocatedLength);
        if (!block) continue;

        LPSTR p = (LPSTR) block;
        LPSTR pend = p + n;

        addBuffer(p, n);

    } // end for all blocks

    return true;
}

void* BuffBlock::addBufferReturnPointer(/*in*/void* p, long length)
{
    if (!addBuffer(p, length))
        return NULL;
    return m_pInsertionBuffer;
}



// CDiffEngine implementation /////////////////////////////////////////////////


CDiffEngine::CDiffEngine()
{
	m_szColorText		= "#888888"; // default colors
	m_szColorBackground = "white";
	m_szColorChanged	= "#FFFFBB";
	m_szColorAdded		= "#BBFFBB";
	m_szColorDeleted	= "#FFBBBB";

	m_szFooter = "<br><font size='-2' color='#BBBBBB'>S.Rodriguez, Aug 2005</font>";
}

// compare f1 (old version) with f2 (new version)
// and build two new copies of those file objects with status on a line by line basis
//
BOOL CDiffEngine::Diff(	/*in*/CFilePartition &f1, /*in*/CFilePartition &f2,
						/*out*/CFilePartition &f1_bis, /*out*/CFilePartition &f2_bis)
{
	f1_bis.SetName( f1.GetName() );
	f2_bis.SetName( f2.GetName() );

	long nbf1Lines = f1.GetNBLines();
	long nbf2Lines = f2.GetNBLines();

	// special empty file case
	if ( nbf1Lines==0 )
	{
		long nLinef2 = 0;
		while ( nLinef2<nbf2Lines )
		{
			f1_bis.AddBlankLine();
			f2_bis.AddString( f2.GetRawLine(nLinef2++), Normal);
		}
		return TRUE;
	}

	long i = 0;
	long nf2CurrentLine = 0;

	while ( i<nbf1Lines )
	{
		// process this line (and possibly update indexes as well)
		long nLinef2 = nf2CurrentLine;

		if ( nLinef2 >= nbf2Lines )
		{
			// it's time to end the game now
			while ( i < nbf1Lines )
			{
				f1_bis.AddString( f1.GetRawLine(i), Deleted);
				f2_bis.AddBlankLine();
				i++;
			}
			break;
		}

		if ( f1.MatchLine(i, f2, /*out*/nLinef2) )
		{
			BOOL bDeleted = FALSE;
			if (nLinef2 > nf2CurrentLine)
			{
				long itmp = i;
				bDeleted = f2.MatchLine(nf2CurrentLine, f1, /*out*/itmp) && (itmp<nbf1Lines);
				if (bDeleted)
				{
					long j = itmp - i;
					while ( j>0 )
					{
						f1_bis.AddString( f1.GetRawLine(i), Deleted);
						f2_bis.AddBlankLine();

						i++;
						j--;
					}
					// please note nf2CurrentLine is not updated

					continue; // jump here to loop iteration
				}
			}

			// matched, so either the lines were identical, or f2 has added one or more lines
			if (nLinef2 > nf2CurrentLine)
			{
				// add blank lines to f1_bis
				long j = nLinef2 - nf2CurrentLine;
				while ( j>0 )
				{
					f1_bis.AddBlankLine();
					f2_bis.AddString( f2.GetRawLine(nLinef2-j), Added );

					j--;
				}
			}

			// exactly matched
			f1_bis.AddString( f1.GetRawLine(i), Normal);
			f2_bis.AddString( f2.GetRawLine(nLinef2), Normal);

			nf2CurrentLine = nLinef2 + 1; // next line in f2

		}
		else
		{
			// this line is not found at all in f2, either it's because it has been changed, or even deleted
			long nLinef1 = i;
			if ( f2.MatchLine(nLinef2, f1, /*out*/nLinef1) )
			{
				// the dual line in f2 can be found in f1, that's because
				// the current line in f1 has been deleted
				f1_bis.AddString( f1.GetRawLine(i), Deleted);
				f2_bis.AddBlankLine();

				// this whole block is flagged as deleted
				if (nLinef1>i+1)
				{
					long j = nLinef1 - (i+1);
					while ( j>0 )
					{
						i++;

						f1_bis.AddString( f1.GetRawLine(i), Deleted);
						f2_bis.AddBlankLine();

						j--;
					}
				}

				// note : nf2CurrentLine is not incremented
			}
			else
			{
				// neither added, nor deleted, so it's flagged as changed
				f1_bis.AddString( f1.GetRawLine(i), Changed);
				f2_bis.AddString( f2.GetRawLine(nLinef2), Changed);

				nf2CurrentLine = nLinef2 + 1; // next line in f2
			}
		}

		i++; // next line in f1
	}

	// are there any remaining lines from f2?
	while ( nf2CurrentLine < nbf2Lines )
	{
		f1_bis.AddBlankLine();
		f2_bis.AddString( f2.GetRawLine(nf2CurrentLine), Added );

		nf2CurrentLine++;
	}


	return TRUE;
}







// build html report
//



void CDiffEngine::SetTitles(CString &szHeader, CString &szFooter)
{
	m_szHeader = szHeader;
	m_szFooter = szFooter;
}

void CDiffEngine::SetColorStyles(CString &szText, CString &szBackground, CString &szChanged, CString &szAdded, CString &szDeleted)
{
	m_szColorText = szText;
	m_szColorBackground = szBackground;
	m_szColorChanged = szChanged;
	m_szColorAdded = szAdded;
	m_szColorDeleted = szDeleted;
}

void CDiffEngine::Serialize(	/*in*/CFilePartition &f1, 
				                /*in*/CFilePartition &f2,
                                /*out*/BuffBlock& buff)
{

	// eval amount of differences between the two files
	int nAdded_f1, nChanged_f1, nDeleted_f1;
	f1.HowManyChanges(/*out*/nAdded_f1, /*out*/nChanged_f1, /*out*/nDeleted_f1);
	int nAdded_f2, nChanged_f2, nDeleted_f2;
	f2.HowManyChanges(/*out*/nAdded_f2, /*out*/nChanged_f2, /*out*/nDeleted_f2);

	int nTotal = nAdded_f1 + nDeleted_f1 + nChanged_f1 + nAdded_f2 + nDeleted_f2;

	if (nTotal==0)
		m_szHeader += "<font size=-1><b>Files are identical.</font><br>";
	else
	{
		TCHAR szTmp[128];
		sprintf(szTmp,"<font size=-1><b>%d changes between the files.</font><br>", nTotal);
		m_szHeader += szTmp;
	}



	// write html header
    buff.addBuffer("<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0 Transitional//EN\">\r\n" \
				"<!-- diff html gen, (c) Stephane Rodriguez - aug 2005 -->\r\n" \
				"<HTML>\r\n" \
				"<HEAD>\r\n" \
				"<TITLE> File Diff </TITLE>\r\n" \
				"<style type='text/css'>\r\n"\
				"<!--\r\n" \
				".N { background-color:white; }\r\n" \
				".C { background-color:");
    buff.addBuffer(m_szColorChanged);
    buff.addBuffer("; }\r\n");
    buff.addBuffer(".A { background-color:");
    buff.addBuffer(m_szColorAdded);
    buff.addBuffer("; }\r\n");
    buff.addBuffer(".D { background-color:");
    buff.addBuffer(m_szColorDeleted);
    buff.addBuffer("; }\r\n-->\r\n" \
				"</style>\r\n" \
				"</HEAD>\r\n" \
				"\r\n" \
				"<BODY BGCOLOR='#FFFFFF'>\r\n" \
				"\r\n");
    buff.addBuffer(m_szHeader);
    buff.addBuffer("<table border=0 bgcolor=0 cellpadding=1 cellspacing=1 width=100%><tr><td>\r\n" \
				"<table width=100% bgcolor=white border=0 cellpadding=0 cellspacing=0>\r\n" \
				"<tr bgColor='#EEEEEE' style='color:0'><td width=50%>old version</td><td width=50%>new version" \
				"&nbsp;&nbsp;&nbsp;(<b style='background-color:");
    buff.addBuffer(m_szColorChanged);
    buff.addBuffer(";width:20'>&nbsp;</b>changed&nbsp;&nbsp;" \
				"<b style='background-color:");
    buff.addBuffer(m_szColorAdded);
    buff.addBuffer(";width:20'>&nbsp;</b>added&nbsp;&nbsp;<b style='background-color:");
    buff.addBuffer(m_szColorDeleted);
    buff.addBuffer(";width:20'>&nbsp;</b>deleted)&nbsp;&nbsp;" \
				"<FORM ACTION='' style='display:inline'><SELECT id='fontoptions' " \
				"onchange='maintable.style.fontSize=this.options[this.selectedIndex].value'>" \
				"<option value='6pt'>6pt<option value='7pt'>7pt<option value='8pt'>8pt<option value='9pt' selected>9pt</SELECT> " \
				"</FORM></td></tr>\r\n" \
				"<tr bgColor='#EEEEEE' style='color:0'><td width=50%><code>");
    buff.addBuffer(f1.GetShortName());
    buff.addBuffer("</code></td><td width=50%><code>");
    buff.addBuffer(f2.GetShortName());
    buff.addBuffer("</code></td></tr>" \
				"</table>\r\n" \
				"</td></tr>\r\n" \
				"</table>\r\n" \
				"\r\n" \
				"<br>\r\n" \
				"\r\n") ;

	long nbLines = f1.GetNBLines();
	if (nbLines == 0)
	{
		buff.addBuffer("<br>empty files");
	}
	else
	{
		buff.addBuffer("<table border=0 bgcolor=0 cellpadding=1 cellspacing=1 width=100%><tr><td>" \
			"<table id='maintable' width=100% bgcolor='");
        buff.addBuffer(m_szColorBackground);
        buff.addBuffer("' cellpadding=0 cellspacing=0 border=0 style='color:");
        buff.addBuffer(m_szColorText);
        buff.addBuffer(";font-family: Courier New, Helvetica, sans-serif; font-size: 8pt'>\r\n");
	}

	
	char *arrStatus[4] = {
		"",
		" class='C'",
		" class='A'",
		" class='D'" };

	TCHAR szLine[16];

	// write content
	//
	for (long i=0; i<nbLines; i++)
	{
		sprintf(szLine, "<b>%d</b>", i);
		buff.addBuffer("<tr><td width=50%");
        buff.addBuffer(arrStatus[ f1.GetStatusLine(i) ]);
        buff.addBuffer(">");
        buff.addBuffer(szLine);
        buff.addBuffer(" ");
        Escape(f1.GetRawLine(i), buff);
        buff.addBuffer("</td>");
		buff.addBuffer("<td width=50%");
        buff.addBuffer(arrStatus[ f2.GetStatusLine(i) ]);
        buff.addBuffer(">");
        buff.addBuffer(szLine);
        buff.addBuffer(" ");
        Escape(f2.GetRawLine(i), buff);
        buff.addBuffer("</td></tr>");
		
	} // for i

	if (nbLines>0)
		buff.addBuffer("</table></td></tr></table>\r\n");

	// write html footer
	buff.addBuffer(m_szFooter);
    buff.addBuffer("</BODY>\r\n" \
		"</HTML>\r\n");
}



void CDiffEngine::Escape(AString &s /*out*/, BuffBlock& o) // a helper aimed to make sure tag symbols are passed as content
{
	long nSize = s.GetLength();
	if (nSize == 0) 
    {
        o.addBuffer("&nbsp;");
        return;
    }

	char c;
	BOOL bIndentation = TRUE;

	for (long i=0; i<nSize; i++)
	{
		c = s.GetAt(i);
		if (bIndentation && (c==' ' || c=='\t'))
		{
			if (c==' ')
				o.addBuffer("&nbsp;");
			else
				o.addBuffer("&nbsp;&nbsp;&nbsp;&nbsp;");
			continue;
		}
		bIndentation = FALSE;

		if (c=='<')
			o.addBuffer("&lt;");
		else if (c=='>')
			o.addBuffer("&gt;");
		else if (c=='&')
			o.addBuffer("&amp;");
		else
			o.addChar(c);
	}
}


BOOL CDiffEngine::ExportAsHtml(/*in*/CString &szFilename, /*in*/BuffBlock& o)
{
	CFile f;
	if ( !f.Open(szFilename, CFile::modeCreate | CFile::modeWrite | CFile::typeBinary) )
	{
		TCHAR szError[MAX_PATH];
		sprintf(szError, "error : cannot create %s\r\n", szFilename.GetBuffer(0) );
		OutputDebugString(szError);
		return FALSE;
	}

    long nbBlocks = o.getBlockCount();
    for (long iBlock = 0; iBlock < nbBlocks; iBlock++)
    {
        long nLength, nAllocatedLength;
        void* block = o.getBlock(iBlock, nLength, nAllocatedLength);
        if (!block || nLength == 0) continue;

	    f.Write(block, nLength);
    } // end for all blocks

	f.Close();

	return TRUE;
}


BOOL CDiffEngine::ExportAsStdout(/*in*/BuffBlock& o)
{
    long nbBlocks = o.getBlockCount();
    for (long iBlock = 0; iBlock < nbBlocks; iBlock++)
    {
        long nLength, nAllocatedLength;
        void* block = o.getBlock(iBlock, nLength, nAllocatedLength);
        if (!block || nLength == 0) continue;

        char* p = new char[1 + nLength];
        if (p)
        {
            memcpy(p, block, nLength);
            p[nLength] = 0;
            printf ("%s", p);
            delete [] p;
        }

    } // end for all blocks


	return TRUE;
}


