


#pragma once

typedef enum {Normal=0, Changed, Added, Deleted} LineStatus;

class CFileLine; // forward declaration
class CFileOptions; // forward declaration
class CFilePartition; // forward declaration


// CFileLine ///////////////////////////////////////////////
//
////////////////////////////////////////////////////////////
// S.Rodriguez - Feb 2003
////////////////////////////////////////////////////////////
//
class CFileLine
{
	// Members
protected:
	AString m_s;
	LineStatus m_status;

	// Constructor
public:
	CFileLine();

	// Accessors
public:
    
	long SetLine(/*in*/AString &s, /*in*/CFileOptions &o); // store string and build token
	void SetLine(/*in*/AString &s, /*in*/LineStatus ls); // store string and status (does not eval token)
	AString GetLine(/*in*/CFileOptions &o);

	void SetStatus(/*in*/LineStatus ls);
	LineStatus GetStatus();
	
	// Helpers
protected:
	AString GetLineWithOptions(/*in*/AString& s,/*in*/CFileOptions &o); // that's where options are implemented
};


typedef CTemplateArray<CFileLine*>		ArrayLines;
typedef CTemplateArray<long>			ArrayLong;
typedef CTemplateObjArray<CString>		ArrayCString;


// CFileOptions /////////////////////////////////////////////////////
//
/////////////////////////////////////////////////////////////////////
// S.Rodriguez - Feb 2003
/////////////////////////////////////////////////////////////////////
//
//
class CFileOptions
{
	// Members
protected:
    typedef std::map<std::string, std::string> mapString2String;
	mapString2String mapOptions;

	// "indent" : should the diff engine see differences in tabs/spaces (default = yes)
	// "case" : should the diff engine see case differences (default = yes)

	// how to use this class :
	//       CFileOptions o;
	//       o.SetOption("indent","yes");

	// Constructor
public:
	CFileOptions();

	// Accessors
public:
	void SetOption(/*in*/std::string& name, /*in*/std::string& value);
        std::string GetOption(/*in*/std::string& name);

	BOOL IsDefaultOptions();
		
	long GetOptionCount();
        void GetOption(long i, std::string& name, std::string& value);

	// Methods
	void Copy(CFileOptions &src);
};








// CFilePartition ///////////////////////////////////////////////////
//
/////////////////////////////////////////////////////////////////////
// S.Rodriguez - Feb 2003
/////////////////////////////////////////////////////////////////////
//
//
class CFilePartition
{

	// Members
protected:
	CString			m_szFilename;
	CFileOptions	m_options;
	ArrayLines		m_arrLines;
	ArrayLong		m_arrTokens;

	// Constructor
public:
	CFilePartition();
	virtual ~CFilePartition();

	// Accessors
public:
	void SetName(/*in*/CString &);
	CString GetName();
    CString GetShortName();

	void SetOptions(/*in*/CFileOptions &options);

	long GetNBLines(); // amount of lines of this file
	AString GetLine(/*in*/long i); // returns an arbitrary line
	AString GetRawLine(/*in*/long i); // returns an arbitrary line (not filtered)
	LineStatus GetStatusLine(/*in*/long i); // returns the status of an arbitrary lin
	long *GetTokens(); // token table

	void HowManyChanges(/*out*/int &nAdded, /*out*/int &nChanged, /*out*/int &nDeleted);


	// Methods
	BOOL PreProcess(/*in*/CString &szFilename, /*in*/CFileOptions &options);
	BOOL PreProcessFolder(/*in*/CString &szFolder, /*in*/CFileOptions &options);

	void AddString(/*in*/AString& s, /*in*/LineStatus ls);
	void AddBlankLine();

	BOOL MatchLine(/*in*/long i1, /*in*/CFilePartition &f2,  /*out*/long &i2);

	void Dump(/*in*/CString &szTitle);


	// Helpers
protected:
	BOOL PreProcessFolder(/*in*/CString &szFolder, /*in*/CFileOptions &options, /*in/out*/long nIndent);
	void AddString(/*in*/AString& s);
};